classdef MfaceAlign
        
    methods (Static)
        
        %   cropFace - Escala, rota i fa el crop d'una cara. La normalitza
        %
        %   Usage:      [caraNormalitzada] = cropFace(imatgeFacial, posUllEsquerre, posUllDret, offset, novesDimensions)        
        %
        %   Arguments:
        %               image - Imatge facial
        %               eye_left - Posici (x,y) ull esquerre
        %               eye_right - Posici (x,y) ull dret
        %               offset_pct - Proporci entre ulls i marges
        %               dest_sz - Noves dimensions (amplada, alada) de la imatge
        %
        %   Returns:
        %               out - Imatge rotada i escalada
        %
        function [ out ] = cropFace( image, eye_left, eye_right, offset_pct, dest_sz )
    
            offset_h = floor(double(offset_pct(1))*dest_sz(1));
            offset_v = floor(double(offset_pct(2))*dest_sz(2));
        
            eye_direction = [eye_right(1) - eye_left(1) eye_right(2) - eye_left(2)];
            rotation = atan2(double(eye_direction(2)),double(eye_direction(1)));

            dist = Utils.distance(eye_left,eye_right);
            reference = dest_sz(1) - 2.0*offset_h;

            scale = double(dist)/double(reference);    
            [image, T2, xdata, ydata] = MfaceAlign.ScaleRotateTranslate(image, rotation, eye_left, 1.0);

            crop_xy = [(eye_left(1)-xdata(1)) - scale*offset_h  (eye_left(2)-ydata(1)) - scale*offset_v];          
            crop_size = [dest_sz(1)*scale dest_sz(2)*scale];
       
            image = imcrop(image,[crop_xy(1) crop_xy(2) crop_size(1) crop_size(2)]);
            out = imresize(image,dest_sz,'Antialiasing',true);
        end
        
        %   cropFaceNoRotate - Escala i fa el crop d'una cara. La normalitza
        %
        %   Usage:      [caraNormalitzada] = cropFaceNoRotate(imatgeFacial, posUllEsquerre, posUllDret, offset, novesDimensions)        
        %
        %   Arguments:
        %               image - Imatge facial
        %               eye_left - Posici (x,y) ull esquerre
        %               eye_right - Posici (x,y) ull dret
        %               offset_pct - Proporci entre ulls i marges
        %               dest_sz - Noves dimensions (amplada, alada) de la imatge
        %
        %   Returns:
        %               out - Imatge rotada i escalada
        %
        function [ out ] = cropFaceNoRotate( image, eye_left, eye_right, offset_pct, dest_sz )
            
            offset_h = floor(double(offset_pct(1))*dest_sz(1));
            offset_v = floor(double(offset_pct(2))*dest_sz(2));
        
            dist = Utils.distance(eye_left,eye_right);
            reference = dest_sz(1) - 2.0*offset_h;

            scale = double(dist)/double(reference);    
        
            crop_xy = [eye_left(1) - scale*offset_h  eye_left(2) - scale*offset_v];          
            crop_size = [dest_sz(1)*scale dest_sz(2)*scale];
       
            image = imcrop(image,[crop_xy(1) crop_xy(2) crop_size(1) crop_size(2)]);
            out = imresize(image,dest_sz,'Antialiasing',true);
        end

        %   rotate - Rota la imatge prenen l'ull esquerre com a centre
        %
        %   Usage:      [out, xdata, ydata] = rotate(imatgeFacial, posUllEsquerre, posUllDret)        
        %
        %   Arguments:
        %               image - Imatge facial
        %               eye_left - Posici (x,y) ull esquerre
        %               eye_right - Posici (x,y) ull dret        
        %
        %   Returns:
        %               out - Imatge rotada
        %               xdata - Vector de dues dimensions, que combinat amb
        %                       ydata especifica la localitzaci espacial
        %                       de la imatge de sortida en 2-D.
        %               ydata - Vector de dues dimensions, que combinat amb
        %                       xdata especifica la localitzaci espacial
        %                       de la imatge de sortida en 2-D.
        %
        function [ out, xdata, ydata ] = rotate(image, eye_left, eye_right)
            
            eye_direction = [eye_right(1) - eye_left(1) eye_right(2) - eye_left(2)];
            rotation = atan2(double(eye_direction(2)),double(eye_direction(1)));

            [image, T2, xdata, ydata] = MfaceAlign.ScaleRotateTranslate(image, rotation, eye_left, 1.0);
            out = image;            
        end
        
        %   rotate2 - Rota la imatge utilitzant un centre concret
        %
        %   Usage:      [out, xdata, ydata] = rotate2(imatgeFacial, centre, angle)        
        %
        %   Arguments:
        %               image - Imatge facial
        %               centre - Centre de rotaci
        %               angle - Angle de rotaci        
        %
        %   Returns:
        %               out - Imatge rotada
        %               xdata - Vector de dues dimensions, que combinat amb
        %                       ydata especifica la localitzaci espacial
        %                       de la imatge de sortida en 2-D.
        %               ydata - Vector de dues dimensions, que combinat amb
        %                       xdata especifica la localitzaci espacial
        %                       de la imatge de sortida en 2-D.
        %
        function [ out, xdata, ydata ] = rotate2(image, centre, angle)
            [out, T, xdata,ydata] = MfaceAlign.ScaleRotateTranslate(image, angle, centre, 1.0);
        end
        
    end

    methods (Static, Access = private)
        
        %   ScaleRotateTranslate - Aplica una transformaci a la imatge
        %
        %   Usage:      [ out, T2, xdata, ydata ] = ScaleRotateTranslate( img, angle, centre, scale )       
        %
        %   Arguments:
        %               img - Imatge facial
        %               angle - Angle de rotaci
        %               centre - Centre de rotaci
        %               scale - Escala de la imatge 
        %
        %   Returns:
        %               out - Imatge transformada.
        %               T2 - Transformaci aplicada.
        %               xdata - Vector de dues dimensions, que combinat amb
        %                       ydata especifica la localitzaci espacial
        %                       de la imatge de sortida en 2-D.
        %               ydata - Vector de dues dimensions, que combinat amb
        %                       xdata especifica la localitzaci espacial
        %                       de la imatge de sortida en 2-D.
        %
        function [ out, T2, xdata, ydata ] = ScaleRotateTranslate( img, angle, centre, scale )
            nx = centre(1);
            ny = centre(2);
    
            x = centre(1);
            y = centre(2);
    
            sx = scale;
            sy = scale;
    
            cosine = cos(angle);
            sine = sin(angle);
    
            a = cosine/sx;
            b = sine/sx;
            c = x-nx*a-ny*b;
            d = -sine/sy;
            e = cosine/sy;
            f = y-nx*d-ny*e;
    
            T2 = maketform('affine',[a d; b e; c f;]);
            [imtrans, xdata, ydata] = imtransform(img,T2,'bicubic');    
            out = imtrans;
        end
    end
end

