classdef MfaceRecognition
        
    methods (Static)
               
        %   loadDatabasePca - Carrega una base de dades per practicar el eigenfaces
        %
        %   Usage:  [w, noms] = loadDatabasePca(newMat, maxImgs,databasePath, databasePathMat)
        %
        %   Arguments:
        %               newMat - Cert si cal crear de zero la base de dades
        %               maxImgs - Nombre d'imatges que es carreguen (per persona)
        %               databasePath - Path d'on es troben les imatges  
        %               databasePathMat - Path d'on es troba el fitxer mat a on es guarden les imatges
        %
        %   Return:
        %               w - matriu d'imatges on cada columna s una imatge
        %               noms - taula on cada cella te el nom de la imatge de la columna de la matriu w         
        %
        function [w, noms] = loadDatabasePca(newMat, maxImgs, databasePath, databasePathMat)            
            if newMat
                v=[];
                noms = [];
                fList = Utils.getSubFolders(databasePath);
                for i=1:size(fList,2)                        
                    iList = Utils.getImages(strcat(databasePath, '\', sprintf('%s',fList{i})));
                    for j=1:maxImgs        
                        % Ha d'estar en escala de grisos                                        
                        a=imread(strcat(databasePath, '\', sprintf('%s',fList{i}), '\', sprintf('%s',iList{j})));
                        
                        if size(a,3)>1
                            a = rgb2gray(a);
                        end
                        
                        v = [v reshape(a,size(a,1)*size(a,2),1)];
                        noms = [noms fList(i)];
                    end                   
                end
                w=uint8(v);
                save(databasePathMat,'w','noms');
            else
                lw = load(databasePathMat,'w');
                lnoms = load(databasePathMat,'noms');
                w = lw.w;
                noms = lnoms.noms;
            end                        
        end
        
        %   loadDatabaseLda - Carrega una base de dades per practicar el fisherfaces
        %
        %   Usage:  [w, noms] = loadDatabaseLda(newMat, maxImgs,databasePath, databasePathMat)
        %
        %   Arguments:
        %               newMat - Cert si cal crear de zero la base de dades
        %               maxImgs - Nombre d'imatges que es carreguen (per persona)
        %               databasePath - Path d'on es troben les imatges  
        %               databasePathMat - Path d'on es troba el fitxer mat a on es guarden les imatges
        %
        %   Return:
        %               w - matriu d'imatges on cada columna s una imatge
        %               noms - taula on cada cella te el nom de la imatge de la columna de la matriu w         
        %
        function [w, noms] = loadDatabaseLda(newMat, maxImgs, databasePath, databasePathMat)            
            if newMat
                v=[];
                noms = [];
                fList = Utils.getSubFolders(databasePath);
                for i=1:size(fList,2)                        
                    iList = Utils.getImages(strcat(databasePath, '\', sprintf('%s',fList{i})));
                    for j=1:maxImgs        
                        % Ha d'estar en escala de grisos                            
                        a=imread(strcat(databasePath, '\', sprintf('%s',fList{i}), '\', sprintf('%s',iList{j})));
                        
                        if size(a,3)>1
                            a = rgb2gray(a);
                        end
                        
                        v = [v reshape(a',size(a,1)*size(a,2),1)];
                        noms = [noms fList(i)];
                    end                   
                end
                w=uint8(v);
                save(databasePathMat,'w','noms');
            else
                lw = load(databasePathMat,'w');
                lnoms = load(databasePathMat,'noms');
                w = lw.w;
                noms = lnoms.noms;
            end                        
        end
        
        %   pcaFaceRecognition - Reconeixement utilitzant eigenfaces
        %
        %   Usage:  [idx, dist, imatge, errorMida] = pcaFaceRecognition(testImage, Nev, trainBdd) 
        %
        %   Arguments:
        %               testImage - Imatge facial a reconeixer
        %               Nev - Nombre d'eigenvectors
        %               trainBdd - Imatges d'entrenament          
        %
        %   Return:
        %               idx - index de la imatge reconeguda dins la base de dades d'entrenament
        %               dist - Distrncia entre imatge de test i la ms propera dins de la base de dades         
        %               imatge - imatge de test
        %               errorMida - Cert si la imatge de test no te les mateixes dimensions que les imatges de la base de dades d'entrenament         
        %        
        function [idx, dist, imatge, errorMida] = pcaFaceRecognition(testImage, Nev, trainBdd)                       
                        
            r = uint8(reshape(testImage,size(testImage,1)*size(testImage,2),1)); 
            errorMida = true;
            idx=[];
            dist=0;
            imatge=[];
                        
            % La mida de les imatges ha de ser la mateixa           
            if size(r,1)==size(trainBdd,1)
                errorMida = false;
                v = trainBdd;           

                N=Nev;                              

                O=uint8(ones(1,size(v,2))); 
                m=uint8(mean(v,2));                 % Mitjana de totes les imatges
                vzm=v-uint8(single(m)*single(O));   % Resta de la mitjana m al conjunt d'entrenament. 


                L=single(vzm)'*single(vzm);
                [V,D]=eig(L);
                V=single(vzm)*V;
                V=V(:,end:-1:end-(N-1));            % Eigenvectors amb els 10 eigenvalues ms grans. 

                cv=zeros(size(v,2),N);

                for i=1:size(v,2);
                    cv(i,:)=single(vzm(:,i))'*V;    
                end

                p=r-m;                              
                s=single(p)'*V;
                z=[];

                for i=1:size(v,2)
                    z=[z,norm(cv(i,:)-s,2)];            
                end

                [a,i]=min(z);
                idx = i;
                dist = a;
                imatge = reshape(v(:,i),size(testImage,1),size(testImage,2));
            end
        end
        
        %   ldaFaceRecognition - Reconeixement utilitzant fisherfaces
        %
        %   Usage:  [idx, dist, imatge, errorMida] = ldaFaceRecognition(testImage, Nev, numperclass, trainBdd) 
        %
        %   Arguments:
        %               testImage - Imatge facial a reconeixer
        %               Nev - Nombre d'eigenvectors
        %               numperclass - Nombre d'imatges per classe
        %               trainBdd - Imatges d'entrenament          
        %
        %   Return:
        %               idx - index de la imatge reconeguda dins la base de dades d'entrenament
        %               dist - Distrncia entre imatge de test i la ms propera dins de la base de dades         
        %               imatge - imatge de test
        %               errorMida - Cert si la imatge de test no te les mateixes dimensions que les imatges de la base de dades d'entrenament         
        %          
        function [idx, dist, imatge, errorMida] = ldaFaceRecognition(testImage, Nev, numperclass, trainBdd)

            idx = [];
            dist = 0;
            imatge = [];
            errorMida = true;
           
            testImage = double(testImage);            
            r = reshape(testImage',size(testImage,1)*size(testImage,2),1);
                                   
            if size(r,1) == size(trainBdd,1)
                errorMida = false;
                [Psi V_KLT V_FLD Proj_fisher] = MfaceRecognition.fisherface(trainBdd, Nev, numperclass);
                [dist, idx] = min(MfaceRecognition.fisher_dist(r,Psi,V_KLT,V_FLD,Proj_fisher));

                imatge = reshape(trainBdd(:,idx),size(testImage,2),size(testImage,1));
                imatge = imatge';
            end
        end
               
        function [w, noms] = loadTestDatabase(newMat, testDatabasePath, databasePathMat)            
            if newMat
                v=[];
                noms = [];
                fList = Utils.getSubFolders(testDatabasePath);
                for i=1:size(fList,2)                        
                    iList = Utils.getImages(strcat(testDatabasePath, '\', sprintf('%s',fList{i})));
                    j = randi([1,size(iList,2)]);
                    
                    % Ha d'estar en escala de grisos                                        
                    rutaImg = strcat(testDatabasePath, '\', sprintf('%s',fList{i}), '\', sprintf('%s',iList{j}));
                    
                    v = [v {rutaImg}];
                    noms = [noms fList(i)];
                    
                end
                w=v;
                save(databasePathMat,'w','noms');
            else
                lw = load(databasePathMat,'w');
                lnoms = load(databasePathMat,'noms');
                w = lw.w;
                noms = lnoms.noms;
            end                        
        end
    end
        
    methods (Static, Access = private)
       
        function [Psi V_KLT V_FLD Proj_fisher] = fisherface(T, Nev, numperclass)
            
            T = double(T);
            numimg = size(T,2);
            numclass = numimg/numperclass; % Nombre de classes 
            numvec = Nev;
            [V_KLT,~,Psi,A] = MfaceRecognition.evectors(T,numvec);
            V_KLT = V_KLT./repmat(sum(V_KLT.^2).^0.5,size(V_KLT,1),1);
            Proj_eig = V_KLT'* A;
            
            % Calcular la mitjana de cada classe dins l'eigenspace
            m_Proj = mean(Proj_eig,2); % Mitjana dins eigenspace
            m = zeros(numvec,numclass); % Mitjana de cada classe dins eigenspace
            %Matrius de disperssi
            Rw = zeros(numvec,numvec); 
            Rb = zeros(numvec,numvec); 

            for i = 1 : numclass
                m(:,i) = mean( ( Proj_eig(:,((i-1)*numperclass+1):i*numperclass) ), 2 )';    
                R  = zeros(numvec,numvec); 
                for j = ( (i-1)*numperclass+1 ) : ( i*numperclass )
                    R = R + (Proj_eig(:,j)-m(:,i))*(Proj_eig(:,j)-m(:,i))';
                end
                Rw = Rw + R; 
                Rb = Rb + (m(:,i)-m_Proj) * (m(:,i)-m_Proj)';
            end

            % Calcular fiser discriminant
            [Vtemp, Dtemp] = eig(Rb,Rw); 
            Dtemp = abs(diag(Dtemp));
            [~, sort_idx] = sort(Dtemp,'descend');            
            int_idx = int8(round(sort_idx));
            V_FLD = Vtemp(:,int_idx);
            
            if numvec >=numclass-1
                V_FLD = V_FLD(:,1:numclass-1);
            else
                V_FLD = V_FLD(:,1:numvec);
            end
            
            V_FLD = V_FLD./repmat(sum(V_FLD.^2).^0.5,size(V_FLD,1),1);

            % Projectar les imatges en l'espai lineal de fisher
            Proj_fisher = V_FLD'*Proj_eig; 
        end
       
        function [Vector,Value,Psi,A] = evectors(T,numvecs)
            
            if nargin ~= 2
                error('Nombre incorrecte d''arguments');
            end;
            numimg = size(T,2);

            % Vector mitjana
            Psi = mean(T,2);

            % Diferncia entre mitjana i cada vector
            A = T - repmat(Psi,1,numimg);

            % Matriu de coovarincia
            L = A'*A;
            [V D] = eig(L);
            
            %Ordenar i convertir els eigenvectora de A'*A a eigenvectors de A*A'
            D = abs(diag(D));
            [Value sort_idx] = sort(D,'descend');
            V = V(:,sort_idx);
            V = V(:,1:numvecs);
            Vector = A*V;
        end
        
        %   fisher_dist - Calcula de la distncia fisher
        %
        %   Usage:  D = fisher_dist(testImage,Psi,V_KLT,V_FLD,Proj_fisher)
        %
        %   Arguments:
        %               testImage - Imatge facial a reconeixer
        %               Psi, V_KLT, V_FLD, PROJ_FISHER - Sortida de la funci fisherface (calcula els eigenvectors generalitzats del conjunt d'imatges d'entrenament)                
        %
        %   Return:
        %               D - Distncia entre projecci de la imatge de test i projecci del conjunt d'entrenament
        %          
        function D = fisher_dist(testImage,Psi,V_KLT,V_FLD,Proj_fisher)

            Proj_test = V_FLD'*V_KLT'*(double(testImage)-Psi);
            D = MfaceRecognition.distance(Proj_test,Proj_fisher);            
        end
        
        %   distance - Calcula de la distncia fisher
        %
        %   Usage:  D = distance(Proj1,Proj2)
        %
        %   Arguments:
        %               Proj1 - Projecci
        %               Proj2 - Projecci
        %
        %   Return:
        %               D - Distncia entre les projeccions
        % 
        function D = distance(Proj1,Proj2)
            
            numimgs1 = size(Proj1,2);
            numimgs2 = size(Proj2,2);
            n = size(Proj1,1);
            D = zeros(numimgs1,numimgs2);
            for i = 1:numimgs2
                D(:,i) = sqrt(sum((Proj1-Proj2(:,i)*ones(1,numimgs1)).^2))'/n;
            end
        end
    end
    
end

